from Products.Archetypes import config
from Products.CMFCore.utils import getToolByName
from AccessControl import ClassSecurityInfo
from Products.Archetypes.Widget import TypesWidget
    
from Products.Archetypes.config import *
from Products.Archetypes.public import *
from Products.Archetypes.exceptions import ReferenceException
from Products.Archetypes.Field import Field,ObjectField
from Products.Archetypes.Widget import ReferenceWidget
from Products.Archetypes.ReferenceEngine import Reference
from Products.Archetypes.Registry import registerField
from Products.Archetypes import public
from Products.generator import i18n
from Products import Archetypes


class BiReferenceWidget(ReferenceWidget):
    _properties = ReferenceWidget._properties.copy()
    _properties.update({
        'macro' : "bireferencewidget",
        })

    security = ClassSecurityInfo()



class BiReferenceField(ReferenceField):
    __implements__ = ObjectField.__implements__

    """A field for creating references between objects.
    Allows for bidirectional relationships by returning references
    created in either direction. References are created on this 
    instance unless set to do the opposite.

    get() returns the list of objects referenced or refencing under the relationship
    set() converts a list of target UIDs into references under the
    relationship associated with this field. References are stored in instance
    unless isbackreference is true

    If no vocabulary is provided by you, one will be assembled based on
    allowed_types.
    """

    _properties = ReferenceField._properties.copy()
    _properties.update({
        'type' : 'bireference',
        'default' : None,
        'widget' : BiReferenceWidget,
        })

    security  = ClassSecurityInfo()

    security.declarePrivate('get')
    def get(self, instance, aslist=False, mimetype=None, **kwargs):
        """get() returns the list of objects referenced under the relationship
        """
        res=instance.getBRefs(relationship=self.relationship)
        res+=instance.getRefs(relationship=self.relationship)

        #if we are indexing this field lets index the titles of the linked objects
        #It's the only thing that makes sense.
        if mimetype == "text/plain":
            return ' '.join([obj.Title() for obj in res])

        #singlevalued ref fields return only the object, not a list,
        #unless explicitely specified by the aslist option
        if not self.multiValued and not aslist:
            if res:
 #               assert len(res) == 1
                res=res[0]
            else:
                res=None
            return res
#        assert type([]) == type(res)
        return res


    security.declarePrivate('set')
    def set(self, instance, value, **kwargs):
        """Mutator.

        ``value`` is a list of UIDs or one UID string to which I will add a
        reference to. None and [] are equal.

        Keyword arguments may be passed directly to addReference(), thereby
        creating properties on the reference objects.
        """
        tool = getToolByName(instance, REFERENCE_CATALOG)
        backtargetUIDs = [ref.sourceUID for ref in
                         tool.getBackReferences(instance, self.relationship)]
        targetUIDs = [ref.targetUID for ref in
                     tool.getReferences(instance, self.relationship)]

        if not self.multiValued and value and type(value) not in (type(()),type([])):
            value = (value,)

        if not value:
            value = ()

        #convertobjects to uids if necessary
        uids=[]
        for v in value:
            if type(v) in (type(''),type(u'')):
                uids.append(v)
            else:
                uids.append(v.UID())

        toobj = lambda uid: tool.lookupObject(uid)
        add = [toobj(v) for v in uids if v and v not in targetUIDs + backtargetUIDs]
        sub = [toobj(t) for t in targetUIDs if t not in uids]
        backsub = [toobj(t) for t in backtargetUIDs if t not in uids]
        
        toreindex = []

        # tweak keyword arguments for addReference
        addRef_kw = kwargs.copy()
        addRef_kw.setdefault('referenceClass', self.referenceClass)
        if addRef_kw.has_key('schema'): del addRef_kw['schema']

            
        for obj in add:
            __traceback_info__ = (instance, obj.UID(), value, targetUIDs)
            # throws IndexError if uid is invalid
            if addRef_kw.get('isBackReference', None):
                src,tgt = obj, instance.UID()
            else:
                tgt,src = obj, instance.UID()
            tool.addReference(src, tgt, self.relationship, **addRef_kw)

        for obj in sub:
            tool.deleteReference(instance.UID(), obj, self.relationship)
        for obj in backsub:
            tool.deleteReference(obj, instance.UID(), self.relationship)

        if self.callStorageOnSet:
            #if this option is set the reference fields's values get written
            #to the storage even if the reference field never use the storage
            #e.g. if i want to store the reference UIDs into an SQL field
            ObjectField.set(self, instance, self.getRaw(instance), **kwargs)

        for obj in sub + backsub + add:
            obj.reindexObject()     

    security.declarePrivate('getRaw') 
    def getRaw(self, instance, aslist=False, **kwargs):
        """Return the list of UIDs referenced under this fields
        relationship
        """
        rc = getToolByName(instance, REFERENCE_CATALOG)
        brains = rc(targetUID=instance.UID(),
                    relationship=self.relationship)
        res = [b.sourceUID for b in brains]
        brains = rc(sourceUID=instance.UID(),
                    relationship=self.relationship)
        res += [b.targetUID for b in brains]
        
        #res=[o.UID() for o in instance.getBRefs()]
        if not self.multiValued and not aslist:
            if res:
                res = res[0]
            else:
                res = None
        return res


#Archetypes.public.__all__=Archetypes.public.__all__+('BiReferenceField',)

#Archetypes.public.BiReferenceField=BiReferenceField

registerField(BiReferenceField,
              title='BiReference',
              description=('Used for storing show both forward and back refs to '
                           'other Archetypes Objects'))
